package Model3;
use strict;
use warnings;

use PDL;
use PDL::Complex;
use PDL::FFT qw(:Func);
use Readonly;
use Benchmark ':hireswallclock';
use Benchmark qw(:all) ;
use Carp;
use Data::Dumper;
use Model;
our @ISA = qw(Model);

Readonly my $POINTS => 40000; #large number of points is more accurate

#Model3.pm is a a object oriented package for calculating the random walk model.
#It takes 4 parameters: p0 - probability of deblocking in a cycle. p1 - probability of incorporating a blocked nucleotide.
#decay - probability of loosing a DNA strand. And, an optional shrinkage factor.
#The main output is a matrix that describes the relative amount of polymerase location in each cycle.




sub new {
        #Create a new object
	my $classname = shift;
	my $self = {};
	$self->{'parameters'} = undef;
	$self->{'CYCLE_END'} = undef;
	bless $self,$classname;
	return $self;
}

sub calculate_matrix {
    #This is the main function of the model. It calculates
    # the polymerase matrix. Each row is a cycle and the
    # columns represent the positions of nucleoties


    #Load parameters
    my $self = shift;
    my ($p0,$p1,$decay,$shrinkage_Factor) = @{$self->{'parameters'}};
    my $maxCycle = $self->{'CYCLE_END'};

    $self->{'shrinkage_Factor'} = $shrinkage_Factor;
            #implictly set the shrinkage factor;

    my $cycle = r2C pdl 0;#Otherwise Cpow provides wrong results;
    my @matrix;

    # Start calculation
    my $exp_i_omegas = Croots(1,$POINTS);
    my $division = Cdiv(
                        $exp_i_omegas * $p0 * $p1
                       ,
                        1-$exp_i_omegas * (1-$p1)
                       )
                 ;
    my $star_basic = 1 - $p0 + $division;
                #star_basic holds the first round of the signal in
                # the frequency domain

    # Calculate the inverse fft for all the rounds.
    foreach my $c (1..$maxCycle) {#check length;
        $cycle++;
        my $star = Cpow ($star_basic,$cycle);
        my $imaginary_star = im $star;
        my $real_star = re $star;
        ifft($real_star,$imaginary_star);
        $real_star = exp(-1*$c*$decay) * $real_star;
        my @tmp = list $real_star; #for some reason it is faster than slicing and .=
        push @matrix,\@tmp;
    }

    # Save and report the results in pdl format
    my $matrix_pdl = pdl \@matrix; 
                            # this is a horizontal rectangular matrix. 
                            # each row is a cycle and the colums are
                            # positions.
    $self->{'matrix'} = $matrix_pdl; 
    return $matrix_pdl; 
            #Extra caution as this retrurn also position #0
}

sub print_matrix_pol {
    my $self = shift;
    my $length = 100;
    if (@_) {
        $length = shift;
    }
    $length--; #pdl first row is 0
    my $matrix_pdl = $self->{'matrix'};
    print $matrix_pdl->slice(":,0:$length");

}



sub launch {
    my $self = shift;
    my $length = $self->{'_sequence_length'};


    if (not exists $self->{'matrix'}) {
        $self->calculate_matrix;
    }
    my $matrix_pol = $self->{'matrix'};
    my $sequence_matrix = $self->get_sequence_matrix;

    my $matrix_pol_sliced = $matrix_pol->slice("1:$length,:");
    my $result =  $matrix_pol_sliced x $sequence_matrix;
    $self->{'_results'} = $result;
    return $result;
}

# sub draw_graph {
# 
#     use GD::Graph;
#     my $self = shift;
#     if (not exists $self->{'_results'}) {
#         carp "can't draw graph without lunching the model";
#         return 0;
#     }
# 
#     my $file='file.png';
#     if (@_) {
#         $file = shift;
#     }
# 
#     my $data_pdl = $self->{'_results'};
#     my @data_array;
#     my $end_cycle = $self->{'CYCLE_END'};
# 
#     #load the data in array format
#     foreach my $nuc (0..3) {
#         my @tmp = list ($data_pdl->slice("($nuc),:"));
#         $data_array[$nuc] = \@tmp;
#     }
# 
# 
# 
#     my $graph  = GD::Graph::mixed->new(1000,600);
#     my @params = (  x_label         => 'Cycle', 
#                     types => ['points', 'points','points','points',
#                               'lines' , 'lines' ,'lines' ,'lines' ,
#                              ],
#                     shadow_depth    => 0,
#                     line_width      =>3,
#                     markers         =>[7,7,7,7],
#                     transparent       => 0,
#                     dclrs => [ qw(blue red green pink 
#                                   blue red green pink 
#                                  )
#                              ],
#                 );
# 
# 
# 
# 
# 
#       $graph->set(@params);
#       my @cycle=(0..($end_cycle-1));
#       my @data_graph=(  \@cycle,
#                   $data_array[0],
#                   $data_array[1],
#                   $data_array[2],
#                   $data_array[3],
#                   $data_array[0],
#                   $data_array[1],
#                   $data_array[2],
#                   $data_array[3],
#                 );
# 
#         my $gd = $graph->plot(\@data_graph) or croak $graph->error; 
#         open IMG, '>', $file or croak $!;
#         binmode IMG;
#         print IMG $gd->png;
# 
# }


sub inverse_polymerase_matrix {

    my $self = shift;
    my $length = $self->{'_sequence_length'};
    if (not defined $length) {
        $length = $self->{'CYCLE_END'};
        $length++;
    }

    my $shrinkage_Factor = $self->{'shrinkage_Factor'};

    if (@_) {
        $length = shift;
    }

    $length--;
    my $full_matrix_pol = $self->{'matrix'};
    my $matrix_pol      = $full_matrix_pol->slice("1:$length,:");

    my $old_method = undef;
    my $new_method = _pseudo_inveret_matrix($matrix_pol,$shrinkage_Factor);

    $self->{'inverse'}->{'old'} = $old_method;
    $self->{'inverse'}->{'new'} = $new_method;

    return ($matrix_pol,$old_method,$new_method);

}


sub _pseudo_inveret_matrix {

    my ($a,$shrinkage_Factor) = @_;
    my ($u,$d,$v) = _svd_new($a);



    if (not defined $shrinkage_Factor) {
        $shrinkage_Factor = 0;
    }

    my $d_diag = $d->diagonal(0,1);

    my $max_D = at($d,0,0);
    my $d_inv_vector = $d_diag /
                                (
                                     $d_diag
                                          **2 
                                 + 
                                  (  $max_D
                                   * $shrinkage_Factor)
                                                      **2
                                );

    my $d_inv = zeroes (dims $d);
    $d_inv->diagonal(0,1) .= $d_inv_vector;

    my @tmp = (dims $a);
    my ($scanDim,$maxDim) = sort @tmp;

    my $tol = ($d->slice('(0),(0)')) * 1e-15 * $maxDim ;

    foreach (0..$scanDim-1) {
        if ($tol>at($d,$_,$_)) {
            set($d,$_,$_,0);
        }
    }


    my $pinv = $v x $d_inv x ($u->transpose);
    return $pinv;
 
}


sub _svd_new {

    my ($matrix) = @_;

    my $matrix_copy = $matrix->transpose;
    my ($cols,$rows) = dims $matrix_copy;
    my @array = list $matrix_copy;
    open (my $TMP,'>','.tmp.del') || croak "can't open tmp file for SVD: $!\n";
        print {$TMP} "$_\n" foreach @array;
    close $TMP;

    my $svd_flag = system ("./svd $rows $cols .tmp.del");



    open (V,'.v.del') || croak "can't open V matrix\n";
    open (D,'.d.del') || croak "can't open D matrix\n";
    open (U,'.u.del') || croak "can't open U matrix\n";

    my @V;
    my @D;
    my @U;
    while (<V>) {
        chomp;
        push @V, $_;
    }
    close V;
    while (<D>) {
        chomp;
        push @D, $_;
    }
    close D;
    while (<U>) {
        chomp;
        push @U, $_;
    }
    close U;

    #unlink('.v.del');
    #unlink('.d.del');
    #unlink('.u.del');

    my $u_pdl = pdl @U;
    my $d_vector_pdl = pdl @D;
    my $v_pdl = pdl @V;

    $u_pdl->reshape($cols,$cols);
    $v_pdl->reshape($cols,$rows);

    my $d_pdl = zeroes ($cols,$cols);
    $d_pdl->diagonal(0,1) .= $d_vector_pdl;

    return ($u_pdl,$d_pdl,$v_pdl);

}

__END__



=head1 NAME
    Model3.pm



=head1 VERSION
    1.0



=head1 SYNOPSIS
    Model3.pm is a a object oriented package for calculating the random walk model.



=head1 METHODS
    new                             create new object
    calculate_matrix                calculate the random walk model.
    print_matrix_pol(<length>)      print the polymerase matrix up to <length>. Default length is 100.
    draw_graph                      
    inverse_polymerase_matrix       inverse the polymerase matrix with two methods.



=head1 DEPENDENCIES
    PDL
    PDL::Complex
    PDL::FFT 
    Readonly
    Benchmark
    Benchmark
    Carp
    Data::Dumper
    Model



=head1 AUTHOR
    Written By Yaniv Erlich <erlich@cshl.edu>

=head1 LISCENSE AND COPYRIGHT
    Copyright © 2008 Cold Spring Harbor Laboratory.
    This is free software.  You may redistribute copies of it under the terms of the GNU General Public License <http://www.gnu.org/licenses/gpl.html>.  There is NO WARRANTY, to the extent permitted by law.

=head1

=cut



# =head1 NAME
# =head1 VERSION
# =head1 SYNOPSIS
# =head1 DESCRIPTION
# =head1 METHODS
# =head1 DIAGNOSTICS
# =head1 CONFIGURATION
# =head1 DEPENDENCIES
# =head1 INCOMPETABILITES
# =head1 BUGS AND LIMITATIONS
# =head1 AUTHOR
#=head1 LISCENSE AND COPYRIGHT





































