#! /usr/bin/env perl

#Copyright 2008 Cold Spring Harbor Laboratory.
# Written by Yaniv Erlich (erlich@cshl.edu)


use strict;
use warnings;
use Getopt::Long;
use Readonly;
use Carp;
use File::Basename;
use File::Copy;

Readonly my $D_GZ_SUFFIX                             => ".gz";
Readonly my $D_GUNZIP                                => 'gunzip';
Readonly my $D_GZIP                                  => 'gzip';
Readonly my $D_SUFFIX_OF_INT_FILES                  => 'int.txt';
            # The suffix illumina uses to mark intensity files in Firecrest
Readonly my $D_SYMBOL                               => '_' ; 
            #The symbol illumina uses to seperate field in files. 
            #For instance '_' in the case of: S_8_0001_seq.txt


main();

sub main {


    print "+load parameters\n";
    my $r_config = read_parameters();

    print "+tile number is:",$r_config->{tile_number},"\n";


    my $int_file = 
    get_filename_in_illumina_format({ path    =>  $r_config->{output_dir},
                                      lane    => $r_config->{lane},
                                      tile    => $r_config->{tile_number},
                                      suffix  => $D_SUFFIX_OF_INT_FILES,
                                      symbol  => $D_SYMBOL});


    print "+int output is $int_file\n";
    $r_config->{output_name} = $int_file;



    print "+merging files\n";
    processing_file($r_config);

}


sub read_parameters {


    my $ipar_file;
    my $pos_file;
    my $output_dir;
    my $tile_number;
    my $lane;
    my $zipped;



    my %config;

    GetOptions( 
                "ipar_file=s"               => \$ipar_file,
                "pos_file=s"                => \$pos_file,
                "output_dir=s"              => \$output_dir,
                "tile_number=i"             => \$tile_number,
                "lane=i"                    => \$lane,
                "zipped"                    =>\$zipped,
               );

     confess "not all parameters were loaded. ",
           "ipar_file: $ipar_file ",
           "pos_file: $pos_file ",
           "output_dir: $output_dir ",
           "tile_number: $tile_number "

         if (
              (not defined $ipar_file)           or
              (not defined $pos_file)            or
              (not defined $output_dir)          or
              (not defined $tile_number)         or 
              (not defined $lane)
            );

    $config{ipar_file}              = $ipar_file;
    $config{pos_file}               = $pos_file;
    $config{output_dir}             = $output_dir;
    $config{tile_number}            = $tile_number;
    $config{zipped}                 = $zipped;
    $config{lane}                   = $lane;

    return \%config;
}






sub processing_file {

        #processing_file is the acutal function that merges the pos files and the intensity files.

        #loading parameters
        my ($r_config) = @_;


        my $ipar_file = $r_config->{ipar_file};
        my $pos_file  = $r_config->{pos_file};
        my $lane      = $r_config->{lane};
        my $tile      = $r_config->{tile_number};
        my $zipped    = $r_config->{zipped};


        my $output_name = $r_config->{output_name} || confess "don't have output name";
        my $output_name_remote = basename $output_name;


       if ($zipped) {
           $ipar_file = _convert_gz_to_text_file($ipar_file);
       } 

        my @lines;

        #giving the developer some data is needed:
        print "+Processing $ipar_file\n";
        print "+pos_file: $pos_file\n";
        print "+output_name: $output_name\n";


        #perpare the files:
        open my $INT,'>', $output_name_remote or confess "can't open $output_name_remote";
            #output file (in the execution node)
        open my $PAR,'<', $ipar_file or confess "can't open $ipar_file";
            # ipar file 
        open my $POS,'<', $pos_file or confess "can't open $pos_file";
            # pos file

        #reading the position file (one per tile). The files are just a list of the cluster positions
        # in the same order of the intensity reads of the ipar file.
        while (<$POS>) {
            chomp;
            my $line = $_;
            $line =~ s/^\s+//;
            my ($x,$y) = split /\s+/, $line;

            #this is annoying but the format pos format has decimal resolution:
            $x = round($x);
            $y = round($y);


            my $header = join "\t", $lane,$tile,$x,$y;
            push @lines, $header;
        }



        #reading the ipar file. Each line is 4 intensity values from a cycle of a given cluster. 
        # new cycle is indicated by '#' sign. Super annoying

        my $i = 0; # cluster counter

        READING_IPAR:
        while (<$PAR>) {
            chomp;
            my $ipar_line = $_;
            if ($ipar_line =~ m/\x23/x) {#this is '# sign'. We are in a new cycle!
                $i = 0; #go back to the first cluster
                next READING_IPAR;
            }

            $lines[$i] .= "\t" . $ipar_line ; #adding another cycle of intensities to the #ith cluster.
            $i++    #go to the next cluster
        }

        # finished to read the ipar file. Now writing
        my $sanity_last;
        foreach my $line_to_write (@lines) {
            my @fields = split /\s+/, $line_to_write;
            my $size = scalar @fields;

            #checking that all lines have the same number of fields. This is a basic sanity check.
            if (defined $sanity_last) {
               if ($sanity_last != $size) {
                  #comapre the number of fields in the new line to the one before.
                  confess "problem. last row had $sanity_last fields and this row has $size. Check yourself: $line_to_write\nDying!!!";
               }
            }

           $sanity_last = $size;
            #writing intensities of the new cluster!!!
           print {$INT} "$line_to_write\n";
        }

        # if the zipped flag is on. Compress the intensity file.
        if ($zipped) {
            $output_name_remote = convert_text_file_to_gz($output_name_remote);
            $output_name = $output_name . $D_GZ_SUFFIX;
        }

        # send it to the right place in the file system.
        copy ($output_name_remote,$output_name) || confess "can't copy $output_name_remote to $output_name:$!\n";
        close $INT;
        close $PAR;
        close $POS;


}



sub _convert_gz_to_text_file {
    my ($gz_file) = @_;
    #my $gz_basename = basename $gz_file;
    #copy($gz_file,"./$gz_basename") or confess "can't copy $gz_file: $!\n";
    #$gz_file = "./$gz_basename";
    system "$D_GUNZIP $gz_file";
    sleep (5);
    $gz_file =~ s/$D_GZ_SUFFIX//;
    return $gz_file;
}

sub convert_text_file_to_gz {
    my ($txt_file) = @_;
    system "$D_GZIP $txt_file";
    sleep (5);
    my $gz_file = join '', $txt_file, $D_GZ_SUFFIX;
    return $gz_file;
}





sub get_filename_in_illumina_format {
    ###############
    #this function generates filenames in solexa format. For example s_4_0002_seq.txt
    #it get in hash table the following keys: path, lane, tile, suffix and symbol.
    #lane and tile are required, default path is './' , default symbol is '_'.
    #path does not need '/' at the end, and symbol is used to concatanate. 
    #for example: get_filename_in_illumina_format({lane => 3, tile => 100, suffix => seq.txt}) 
    #will give ./s_3_0100_seq.txt
    ###############

    my ($arg_ref) = @_;

    my $path      = $arg_ref->{path} || '.';
    my $lane      = $arg_ref->{lane};
    my $tile      = $arg_ref->{tile};
    my $suffix    = $arg_ref->{suffix};
    my $symbol    = $arg_ref->{symbol} || '_';

    my @results;




        #generating /path/s_<lane>_000<tile>_<suffix>. like ./s_8_0001_seq.txt
        my $tile_with_zeros = sprintf "%04d" , $tile ;
        my $file = join $symbol, 's', $lane, $tile_with_zeros;
        if (defined $suffix) {
            $file = join $symbol, $file, $suffix;
        }

        my $file_full = join '/', $path, $file;

    return $file_full;
}




sub round {
    my($number) = shift;
    return int($number + .5);
}

