#! /usr/bin/env perl

#Copyright 2008 Cold Spring Harbor Laboratory.
# Written by Yaniv Erlich (erlich@cshl.edu)

use strict;
use warnings;
use PDL;
use Readonly;
use Storable;
use Carp;
use Archive::Tar;
use Cwd 'abs_path';
use File::Copy;
use File::Temp qw (tempfile);
use List::Util qw (shuffle);
use Getopt::Long;


Readonly my $D_GZ_SUFFIX                             => ".gz";
Readonly my $D_GUNZIP                                => 'gunzip';
Readonly my $D_GZIP                                  => 'gzip';
Readonly my $D_SUFFIX_OUTPUT                         => '.int.filtered';
Readonly my $D_TMPELATE                              => 'intensity_file_to_filtered_XXXX';


Readonly our $D_SUFFIX_OF_INT_FILES                  => 'int.txt';
            # The suffix illumina uses to mark intensity files in Firecrest

Readonly our $D_SYMBOL                               => '_' ; 
            #The symbol illumina uses to seperate field in files. 
            #For instance '_' in the case of: S_8_0001_seq.txt


#pre_process_intensities.pl --intensity_file_to_read lala.txt --pos_to_seq baba.hash --output_dir ./ 

main();


sub main {

    print "+load parameters\n";
    my $r_config = read_parameters();

    print "+tile number is:",$r_config->{tile_number},"\n";

    my $lines_to_read = report_number_of_lines_to_read($r_config);
    print "+total lines to read:$lines_to_read\n";


    print "+processing sequences\n";

    my ($outputfile,$number_of_lines) = take_snapshot_int($r_config , $lines_to_read);

    print "$outputfile\n";
    print "+Done. Found $number_of_lines intensity lines in ",$r_config->{intensity_file_to_read},"\n";


}




sub read_parameters {

    #my $self = shift;

    my $intensity_file_to_read;
    my $pos_to_seq;
    my $output_dir;
    my $outputfile;
    my $tile_number;

    my %config;

    GetOptions( 
                "intensity_file_to_read=s"      => \$intensity_file_to_read,
                "pos_to_seq=s"                  => \$pos_to_seq,
                "output_dir=s"                  => \$output_dir,
                "tile_number=s"                 => \$tile_number,
               );


     croak "not all parameters were loaded. ",
           "intensity_file_to_read: $intensity_file_to_read ",
           "pos_to_seq: $pos_to_seq ",
           "output_dir: $output_dir ",
         if (
              (not defined $intensity_file_to_read)                 or
              (not defined $pos_to_seq)                             or
              (not defined $output_dir)
            );

    $config{intensity_file_to_read} = $intensity_file_to_read;
    $config{pos_to_seq}             = $pos_to_seq;
    $config{output_dir}             = $output_dir;
    $config{tile_number}            = $tile_number;

    return \%config;
}


sub report_number_of_lines_to_read {
    my ($r_config) = @_;

    my $pos_to_seq = $r_config->{pos_to_seq};
    my $tile_number = $r_config->{tile_number};

    my $r_pos_to_seq = retrieve ($pos_to_seq);
    my $lines_to_read = 0;


    foreach my $entry (keys %$r_pos_to_seq) {
        my ($lane,$tile,$x,$y) = split /_/ , $entry;
        if ($tile == $tile_number) {
            $lines_to_read++;
        }
    }
    return $lines_to_read;

}

sub take_snapshot_int {
    my ($arg_ref, $lines_to_read) = @_;

    my $pos_to_seq              =   $arg_ref->{'pos_to_seq'}; #this is hash file
    my $output_dir              =   $arg_ref->{'output_dir'};
    my $int_file_to_read        =   $arg_ref->{'intensity_file_to_read'}; #should be fullname`

    my $outputfile              =   $output_dir. '/'. $int_file_to_read . $D_SUFFIX_OUTPUT;


    my $r_pos_to_seq = retrieve($pos_to_seq) or 
        confess "can't retrieve $pos_to_seq: $!";




    my $r_intesity_lines = _filter_intensities (
                                                $int_file_to_read,
                                                $r_pos_to_seq,
                                                $lines_to_read,
                                               );

    my $tmp_template = $D_TMPELATE;
    my ($FH_OUT,$fname) = tempfile($tmp_template)
             or croak "can't open tempfile\n";


    print {$FH_OUT} $_ foreach @$r_intesity_lines;

    close $FH_OUT
        or confess "can't close $outputfile: $!\n";


    move($fname,$outputfile)
        or confess "can't move $fname to $outputfile: $!\n";


    return ($outputfile,scalar @$r_intesity_lines);

}

sub _filter_intensities {

    my ($file,$r_pos_to_seq,$lines_to_read) = @_;

    my $gz_flag = 0;
    my @intensity_lines;

    if ($file =~m/$D_GZ_SUFFIX/) {
        $file = _convert_gz_to_text_file($file);
        $gz_flag = 1;
    }

    open my $FH_IN, '<', $file
        or confess "can't open $file: $!\n";

    while (<$FH_IN>) {

        my $line = $_;
        my @list=split /\s+/ , $line;

        my ($lane,$tile,$x,$y) = @list[0..3];

        my $name    = join '_',$lane,$tile,$x,$y;


        if (exists $r_pos_to_seq->{$name}) {
            push @intensity_lines, $line;
            $lines_to_read--;
            delete $r_pos_to_seq->{$name};
        }

    }
    close $FH_IN
        or confess "can't close $file: $!\n";

    if ($gz_flag == 1) {
        unlink $file || confess "can't clean $file: $!";
    }

    print "+number of lines that remained to read:$lines_to_read\n";

    @intensity_lines = shuffle @intensity_lines;

    return \@intensity_lines;

}



sub _convert_gz_to_text_file {
    my ($gz_file) = @_;
    my $gz_basename = basename $gz_file;
    copy($gz_file,"/tmp/$gz_basename") or confess "can't copy $gz_file: $!\n";
    $gz_file = "/tmp/$gz_basename";
    system "$D_GUNZIP $gz_file";
    sleep (5);
    $gz_file =~ s/$D_GZ_SUFFIX//;
    return $gz_file;
}


