#Copyright 2008 Cold Spring Harbor Laboratory.
# Written by Yaniv Erlich (erlich@cshl.edu)

# This package generates several methods for arranging job arrays before submitting on SGE.
# The main output is creation of an ad-hoc perl script. Once the script submitted (by a different package)
# it brings all required files to the execution node, calculates the coordinate from the job number, and executes the job. 

package SGE_API;

use strict;
use warnings;
use Readonly;
use List::MoreUtils qw (any);
use Carp;
use Cwd;
use File::Temp qw(tempfile);
use ExtraUtils qw(timestamp_file);
use File::Basename;
use Cwd 'abs_path';
use Permute qw(permute);
use Storable;

Readonly my $D_FOLDER           => '.'                                  ;
Readonly my $D_FILENAME         => 'sge_script'                         ;
Readonly my $D_REMOTE_FOLDER    => '/tmp'                               ;
Readonly my $D_PERL_VERSION     => '/usr/bin/env perl'                  ;
Readonly my $DATA_FILE_TEMPLATE => './dataFile.tmp'                     ;



sub test {
# #     my $api = SGE_API->new;
#     $api->add_required_file({refer=>'file1',
#                              place=>'local',
#                              name=>'delme.pl.<_>'});
#    $api->add_command('echo <X>');
#     my @start=(0.95,0.95);
#     my @stop=(1,1);
#     my @gap=(0.005,0.005);
# 
#     my ($c,$r_l) =  $api->set_range(\@start,\@stop,\@gap);
#     print $c,"\n";
#     print "@$r_l ","\n";
# 
#    print $api->compile,"\n";
#    $api->check_integrity;
#     $api->bye;

}


sub bye {

    my $self = shift;
    my $file_range = $self->get_range_file_name;
    unlink $file_range
        or confess "can't clean $file_range\n";

}

sub new {
        #Create a new object
	my $classname = shift;
	my $self = {};
	bless $self,$classname;
        my ($filehandle,$filename) = _open_new_file();
        $self->{meta}->{filehandle} = $filehandle;
        $self->{meta}->{filename}   = $filename;
        $self->{data}->{home_dir}   = cwd();
        $self->{data}->{remote_dir} = $D_REMOTE_FOLDER;

        $self->add_wildcards;
	return $self;
}


sub set_range {

    Readonly my $SEP => '_';

    my ($self,$r_start,$r_end,$r_gap) = @_;
    my @permuatitons;
    my %vec_of;
    my $counter=0;
    my @lengths;

    $vec_of{verify} = 1;
    $vec_of{sep} = $SEP;

    confess "set_range error\n"
        if any {not defined $_} $r_start, $r_end;

    if (not defined $r_gap) {
        my @a;
        foreach (@$r_start) {push @a , 1}
        $r_gap = \@a;
    }

    if ((scalar @$r_start != scalar @$r_end) or 
        (scalar @$r_end   != scalar @$r_gap))  {
        confess "set_range error - lengths of arrays must be the same";
    }

    foreach my $dimenstion (0..scalar @$r_start-1) {
        my $start = $r_start->[$dimenstion];
        my $stop  = $r_end->[$dimenstion];
        my $gap   = $r_gap->[$dimenstion];

        my @tmp;
        for (my $i=$start; $i<=$stop; $i+=$gap) {
            push @tmp, $i;
        }
        push @lengths, scalar @tmp;
        push @permuatitons, \@tmp;
    }

    @permuatitons = permute(@permuatitons);
    foreach my $p (@permuatitons) {
        $counter++;
        my $vec = join ($SEP, @$p);
        #print "$counter\t","$vec\n";
        $vec_of{$counter} = $vec;
    }

    confess "set_range error - counter is zero!\n",
          "Parameters:\n",
          "starts: @$r_start\n",
          "stops:  @$r_end\n",
          "gaps:   @$r_gap\n"
        if (not $counter);



    my $filename = $DATA_FILE_TEMPLATE . _get_suffix();
    store (\%vec_of, $filename)
        or confess "set_range error - cant store file\n";

    my $full_filename = abs_path $filename;
    $self->add_required_file ({refer=>'hashFile',
                               name=>$full_filename,
                               place=>'local'});

    $self->set_range_flag;
    $self->set_range_file_name($full_filename);

    #counter is total number of permutations
    return ($counter,\@lengths);
}






sub set_range_file_name {
    my ($self,$name) = @_;
    confess "set_range_file_name error\n"
        if (not defined $name);

    $self->{meta}->{rangefile} = $name;
    return $self->{meta}->{rangefile};
}

sub get_range_file_name {
    my $self = shift;
    return $self->{meta}->{rangefile};
}


sub add_wildcards {
    my $self = shift;
    $self->add_required_file({refer=>'_',
                              place=>'local',
                              name=>'$TASK_ID'
                            });

    $self->add_required_file({refer=>'X',
                              place=>'local',
                              name=>'$NUM_VEC'
                            });

    return 1;
}

sub remove_wildcard {
    my $self = shift;
    delete $self->{data}->{required_files}->{'_'};
    delete $self->{data}->{required_files}->{'X'};
    return 1;
}

sub add_required_file {

    my $self = shift;
    if (@_) {
        my ($arg_ref) = @_;

        my $source = undef;
        my $dest_exp = undef;
        my $dest_link = undef;

        my $refer = $arg_ref->{'refer'};

        my $locality_flag = $arg_ref->{'place'} || 'local';
        confess "add_required_file error\n"
            if (
                  (not (($locality_flag eq 'distal') or ($locality_flag eq 'local')))
                 or
                  (not defined $arg_ref->{name} or not defined $arg_ref->{'refer'})
               );

        my $name = $arg_ref->{'name'};
        $name = $self->_parser($name);
        if ($locality_flag eq 'local') {
            $source = $name;
            $dest_link = basename $name;
        }
        else {
            $dest_exp = $name;
            $dest_link = $dest_exp;
        }
        $self->{data}->{required_files}->{$refer}->{source}          = $source;
        $self->{data}->{required_files}->{$refer}->{remote}          = $dest_link;
        $self->{data}->{required_files}->{$refer}->{remote_expected} = $dest_exp;

        return scalar keys %{$self->{data}->{required_files}->{$refer}};
    }

    return;
}

sub add_command {

    my $self = shift;
    if (@_) {
        my ($command,$flag_resolve_filename) = @_;

        $flag_resolve_filename = 1
            if (not defined $flag_resolve_filename);


        if ($flag_resolve_filename == 1) {
            $command = $self->_parser($command);
        }

        if ($command =~ m/;/) {
            carp "add_command problem: you don't need to add ';' at commands\n";
        }

        $command = 'system ("'  .
                    $command    . 
                    '")'
                    ;

        push (@{$self->{data}->{command_list}},$command);
        return scalar @{$self->{data}->{command_list}};
    }
    return;
}


sub _parser {

    my ($self, $input) = @_;
    my $output = $input;
            while ($input =~ m/<(\w*)>/g) {
                my $change_me = $1;
                my $changed = $self->{data}->{required_files}->{$change_me}->{remote} ||
                    carp "can't find any matching to $change_me in _parser function\n";
                $output =~ s/<$change_me>/$changed/;
            }
    return $output;
}


sub check_integrity {

    my $self = shift;
    if (not exists $self->{meta}->{flags}->{compile}) {
        confess "must compile before checking integrity\n";
    }

    my $file = get_filename($self);
    my $response = `$D_PERL_VERSION -c $file 2>&1`;
    if (not $response =~ m/syntax OK/) {
        confess "check_integrity file:$file failed in syntax test: $response\n";
    }
    return 1;

}

# sub change_homedir {
# 
#     my $self = shift;
#     if (@_) {
#         $self->{data}->{home_dir} = shift;
#         return $self->{data}->{home_dir};
#     }
#     return;
# }

sub set_range_flag {
    my $self = shift;
    $self->{meta}->{flags}->{set_range}=1;
    return 1;
}
sub get_range_flag {
   my $self = shift;
   return $self->{meta}->{flags}->{set_range};
}


sub compile {
    my $self = shift;
    $self->remove_wildcard;

    confess "can't complie before setting range\n"
        if (not defined $self->get_range_flag);

    my $fh = get_filehandle($self);
    my $filename = get_filename($self);
    _write_headers($fh,$self->{data}->{remote_dir},$self->{data}->{home_dir});
    _print_hash_files($self->{data}->{required_files},$fh);
    _print_commands($self->{data}->{command_list},$fh);
    _print_epilogue($fh);

    close $fh;
    chmod oct '0777',$filename
        or confess "can't chmod $filename in compile function\n";

    $self->{meta}->{flags}->{compile} = 1;
    return $filename;
}


sub change_remotedir {

    my $self = shift;
    if (@_) {
        $self->{data}->{remote_dir} = shift;
        return $self->{data}->{remote_dir};
    }
    return;
}

sub get_filename {

    my $self = shift;
    return $self->{meta}->{filename};
}

sub get_filehandle {

    my $self = shift;
    return $self->{meta}->{filehandle};

}

sub _open_new_file {

    my $suffix = _get_suffix();

    if (not -e $D_FOLDER) {
        mkdir ($D_FOLDER) 
            or confess "_open_new_file error\n";
    }

    my $filename = "$D_FOLDER/$D_FILENAME"."_$suffix.pl";
    while (-e $filename) {
        my ($name,$path,$suffix) = fileparse $filename, qr/\d*/;
        if ($suffix) {$suffix++} else {$suffix=1};
        $name =~ s/\.$//;
        $filename = $path . "/". $name . ".$suffix";
    }

    open my $FH, '>', $filename 
        || confess "can't open $filename for writing in _open_new_file function: $!\n";



    $filename = abs_path($filename);
    return ($FH,$filename);

}

sub _get_suffix {

    my ($sec,$min,$hour,$day, $month,$year) = localtime(time) ;
    $year  += 1900;
    $month += 1;
    my $suffix = sprintf("%04d_%02d_%02d_%02d_%02d_%02d", $year,$month,$day,$hour,$min,$sec);
    return $suffix;

}

sub _write_headers {

    my ($FH,$remote_dir,$source_dir) = @_;

    print {$FH} "#! $D_PERL_VERSION
                    use strict;
                    use warnings;
                    use File::Copy;
                    use Readonly;
                    use Storable;
                    use Carp;

                    my \$TASK_ID = shift;

                    Readonly my \$REMOTE_DIR => \'$remote_dir\';
                    Readonly my \$SOURCE_DIR => \'$source_dir\';
                    print STDOUT \"+I am alive: \$TASK_ID\\n\";
                ",
                '
                    sub required_files {
                        my ($r_hash_files) = @_;
                        foreach my $file (keys %{$r_hash_files}) {
                            my $file_source          = $r_hash_files->{$file}->{source};
                            my $file_dest            = $r_hash_files->{$file}->{remote};
                            my $file_dest_expected   = $r_hash_files->{$file}->{remote_expected} || $file_dest;

                            if (-e $file_source) {
                                copy($file_source,$file_dest)
                                    or die "can not copy $file_source to $file_dest\n";
                                chmod oct \'0777\', $file_dest
                                    or die "can not chmod $file_dest: $!\n";
                            }

                            if (not -e $file_dest) {
                                die "cant fount $file_dest\n";
                            }
                        }
                    }
                ',

                ' sub convert_task {
                    Readonly my $DATA_FILE => \'./dataFile\';
                    my ($convert_me,$r_hash_files) = @_;
                    my $r_hash_numbers = retrieve ($r_hash_files->{hashFile}->{remote}) 
                        or confess "cant open dataFile\n";

                    confess "cant verify hash\n"
                        if (not exists $r_hash_numbers->{verify});

                    my $seperator = $r_hash_numbers->{sep} || "_";

                    my $numbers = $r_hash_numbers->{$convert_me}
                        or confess "cant convert $convert_me\n";
                    $numbers =~ s/$seperator/ /g;
                    return $numbers;
                }
                '

}



sub _print_hash_files {

    my ($r_hash, $fh) = @_;
    confess "_print_hash_files error\n" 
        if (not defined $r_hash);

    $fh = &STDOUT if (not defined $fh);



    #this code only prints the file data sctructure to get 
    # the hash table that adapts between $TASK_ID to $NUM_VEC 
    print {$fh} "\n#hash of NUM_VEC:\n";
    print {$fh} 'my $r_hash_NUM_VEC=({';
    GET_NUM_VEC:
    foreach my $file (keys %{$r_hash}) {
        next GET_NUM_VEC if ($file ne 'hashFile');
        my $source = $r_hash->{$file}->{source} || 'undef';
        my $dest_f = $r_hash->{$file}->{remote} || 'undef';
        my $dest_exp = $r_hash->{$file}->{remote_expected} || 'undef';
        print {$fh} "\t$file=>{source=>\"$source\",remote=>\"$dest_f\",remote_expected=>\"$dest_exp\"},\n";
    }
    print {$fh} "});\n";
    delete ($r_hash->{hashFile}); #not to print twice the same file in two hashes.
    print {$fh} '#First - Solve $NUM_VEC',"\n",
                'required_files ($r_hash_NUM_VEC);',"\n",
                'my $NUM_VEC = convert_task($TASK_ID, $r_hash_NUM_VEC);',"\n\n\n";


    #hash
    print {$fh} "\n#hash files:\n";
    print {$fh} 'my $r_hash_files=({';

    foreach my $file (keys %{$r_hash}) {
        my $source = $r_hash->{$file}->{source} || 'undef';
        my $dest_f = $r_hash->{$file}->{remote} || 'undef';
        my $dest_exp = $r_hash->{$file}->{remote_expected} || 'undef';
        print {$fh} "\t$file=>{source=>\"$source\",remote=>\"$dest_f\",remote_expected=>\"$dest_exp\"},\n";
    }
    print {$fh} "});\n";







}

sub _print_commands {

    my ($r_command_list,$fh) = @_;
    confess "_print_commands error\n" 
        if (not defined $r_command_list);

    $fh = &STDOUT if (not defined $fh);

    print {$fh} "\n\n\n\n";
    print {$fh} "#commands:\n";
    print {$fh} 'required_files ($r_hash_files);',"\n";


    foreach my $command (@{$r_command_list}) {
        print {$fh} "$command;\n";
    }
}

sub _print_epilogue {

    my ($fh) = @_;
    $fh = &STDOUT if (not defined $fh);

    print {$fh} "#epilouge:\n";
    print {$fh} "#\n\n\n";

}




1;
