#Copyright 2008 Cold Spring Harbor Laboratory.
# Written by Yaniv Erlich (erlich@cshl.edu)

#This package samples Solexa data (intensity values and sequences) and creates the training set.

package Snapshot;
require Exporter;

our @ISA    = ("Exporter");
our @EXPORT = qw(quick $D_SUFFIX_OF_INT_FILES $D_SYMBOL $D_FOLDER_TRAIN_TRAIN);
our $VERSION = 1.00;


use strict;
use warnings;
use Readonly;
use List::Util qw (shuffle);
use List::MoreUtils qw (any);
use Storable;
use Carp;
use FindBin;
use lib $FindBin::Bin;
use File::Basename;
use File::Copy;



Readonly our $D_FOLDER_TRAIN                         => 'Training_set_files';
Readonly my  $BLAT_MAXIMUM_HEADER_LINE               => '500';

Readonly my  $D_NAME_OF_UNIFIED_SAMPLE_SEQ           => 'unified_seq_sample.txt';
Readonly my  $D_NAME_OF_UNIFIED_SAMPLE_SEQ_FOR_BLAT  => 'unified_seq_sample_for_blat.fa';
Readonly our $D_SUFFIX_OF_INT_FILES                  => 'int.txt';
            # The suffix illumina uses to mark intensity files in Firecrest

Readonly my $D_SUFFIX_OF_SEQ_FILES                  => 'seq.txt'; 
            # The suffix illumina uses to mark sequence files in Bustard
Readonly our $D_SYMBOL                               => '_' ; 
            #The symbol illumina uses to seperate field in files. 
            #For instance '_' in the case of: S_8_0001_seq.txt



#^^^^^^^^^^^^^^^^^Headers^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^


############################################ MAIN FUNCTIONS ######################################
sub quick_snapshot_sequences {

    my ($arg_ref) = @_;

    my $verbose           = $arg_ref->{verbose};            #verbosity
    my $path_to_seq_files = $arg_ref->{path_to_seq_files};
    my $trim              = $arg_ref->{trim};
    my $output_dir        = $arg_ref->{output_dir};
    my $back              = $arg_ref->{maximum_lines_for_training};
    my $lane              = $arg_ref->{lane};
    my $r_tiles           = $arg_ref->{tiles};

    if (not -e "$output_dir/$D_FOLDER_TRAIN") {
        mkdir "$output_dir/$D_FOLDER_TRAIN" or confess "can't make folder: $D_FOLDER_TRAIN $!";
    }

    $output_dir = "$output_dir/$D_FOLDER_TRAIN";

    my @seq_files_fullname = 
        get_filename_in_illumina_format({path   => $path_to_seq_files,
                                        lane    => $lane,
                                        tiles   => $r_tiles,
                                        suffix  => $D_SUFFIX_OF_SEQ_FILES,
                                        symbol  => $D_SYMBOL});


    my ($uni_seq_file,$uni_int_file) = 
        take_snapshot_seq({
                            verbose=>$verbose,
                            back=> $back,
                            output_dir=>$output_dir,
                            seq_files=>\@seq_files_fullname,

                          });

    my ($uni_seq_to_blat_file, $unique_seq) = 
        prepare_to_blat({   
                         verbose=>$verbose,
                         unified_seq_sample=>$uni_seq_file,
                         output_dir=>$output_dir,
                         trim=>$trim,
                         number_of_maximum_lines=>$back,
                        });

    return $uni_seq_to_blat_file;

}





############################# TAKE A SANPSHOT #######################################################3
sub take_snapshot_seq {
    #this function samples random <back> lines from the training library - only the training files
    # it returns the filename of the sampled sequences
    # NOTICE: SEQUENCES WITH 'N' OR '.' ARE DISCARDED!

    my ($arg_ref) = shift;

    my $r_seq_files_to_read     =   $arg_ref->{'seq_files'};            #should be fullname files
    my $output_dir              =   $arg_ref->{'output_dir'} || '.';
    my $number_of_lines_back    =   $arg_ref->{'back'};                 #number of maximum lines to getback
    my $verbose                 =   $arg_ref->{'verbose'};


    # the output file are defined as constant:
    my $outputfile = "$output_dir/$D_NAME_OF_UNIFIED_SAMPLE_SEQ";

    print "Taking sanpshot of sequences from the training library\n",
          "Reading training library\n"
        if ($verbose);


    #total lines - how many lines do we have in total in the trainings library (on the tiles that were given).
    #r_valid_places - a reference for array with location of lines with no Ns.
    my ($total_lines,$r_valid_places) = 
             total_number_of_lines_for_training({
                                                 files     => $r_seq_files_to_read,
                                                 verbose   => $verbose,
                                                 });

    #Scream if the requested lines back are more than the lines in the all library:
    confess "The number of total reads without Ns in the training library is ", scalar @$r_valid_places,
            " < $number_of_lines_back\n"
          if (scalar (@$r_valid_places) < $number_of_lines_back);


    # a reference to hash that the keys are line numbers (as read from the first file).
    # if the first file has 10000 lines and the second has 7000, then 11,000 referes to the 
    # 1000th line in the second file. The value is just 1 as it is used to check existsness.

    # the function _genereate_random_sample_from return unique random <$number_of_lines_back> line numbers
    # from the r_valid_places
    my $r_is_line_there     = 
            _genereate_random_sample_from($r_valid_places,
                                          $number_of_lines_back,
                                         );

    #after we have the random lines numbers lets actual read them.

        #open the file for wtiring:
        open my $FH_OUT, '>', $outputfile
            or confess "can't open $outputfile: $!\n";

        _sample($r_seq_files_to_read,
                $r_is_line_there,
                $FH_OUT,
                $verbose,
               );

        close $FH_OUT
            or confess "can't close $outputfile: $!\n";


    print "Generated sample of sequences for training: $outputfile\n"
        if ($verbose);

    #return the fullnames of the written file
    return $outputfile;
}



sub _genereate_random_sample_from {
    my ($r_valid_places,$number_of_lines_back) = @_;

    #r_valid_places = reference to array of lines without Ns.
    #number_of_lines_back = how many to sample out of them
    my @all_possible_lines = shuffle @$r_valid_places;
    my %is_line_there = map {$_ => '1'} @all_possible_lines[0..$number_of_lines_back-1];
    #return a hash with the line numbers as keys
    return \%is_line_there;
}



sub total_number_of_lines_for_training {
    #get list of files and return the total number of lines
    # and valid line (line without Ns)

    my ($arg_ref)           = shift;
    my $r_files_to_read     = $arg_ref->{'files'};
    my $verbose             = $arg_ref->{'verbose'}    || 0;


    my $total_lines         = 0;#counter
    my @valid_lines         = undef;


    confess "<files> should be array refernce\n"
        if (ref $r_files_to_read ne 'ARRAY');

    READING_FILES:
    foreach my $file (@$r_files_to_read) {
        open my $FH, '<', $file
            or confess "can't open $file: $!\n";
        while (<$FH>) {
            $total_lines++;

            push @valid_lines, $total_lines
                if ( (not m/\./) and (not m/N/) );
        }
        close $FH
            or confess "can't close $file: $!\n";
    } #end of READING_FILES

    print "Found: $total_lines lines in total\n",
          "Lines without Ns: ",scalar @valid_lines, "\n"
        if ($verbose);


    return ($total_lines,\@valid_lines);
}



sub _sample {
    my ($r_files,$r_is_line_there,$FH_OUT,$verbose) = @_;

    #r_files = ref to array of files to read
    #r_is_line_there = hash function with line numbers
    #                  to sample as keys
    #FH_OUT = file handle for output data

    my $counter = 0;
    my @lines;

    foreach my $file (@$r_files) {

        open my $FH_IN, '<', $file
            or confess "can't open $file: $!\n";

        print "Processing $file\n"
            if ($verbose);

        while (<$FH_IN>) {
            $counter++;
            if (exists $r_is_line_there->{$counter}) {
                #push @lines, $_;
                 print {$FH_OUT} "$_"
                    or confess "can't print to file: $!";
            }
        }

        close $FH_IN
            or confess "can't close $file: $!\n";


    }

    #obsolete area:
        #@lines = shuffle @lines;
        # shuffle is for generating random order of training entries
        # the problem that solexa intensity files and seqeunce files
        #  are ordered by signal intensity.
    #
    return 1;
}





############################# Prepare to blat #################################################################################


sub prepare_to_blat {
    #############################
    #this function create a dataset for training seqeunce. The sequence are trimmed 
    #and collapsed to boost blat search. The output file is in fasta format for blat.
    #  The header of each sequence is sarting with leading '>' and ';' seperates 
    # between the location of entries with the same sequence. 
    # It returns the path/filename of the output sequences and number of unique sequences. 

    # NOTICE: SEQUENCES WITH 'N' OR '.' ARE DISCARDED!

    #the inputs of the function are in a hash table with the following keys: unified_seq_sample,
    # trim, output_dir,number_of_maximum_lines and verbose. Only unified_seq_sample is mandatory.
    # unified_seq_sample - a fullname of tab delimited file with cluster position as the first four colums
    # trim - number of bases to keep for training from the starting point of the sequence
    ##############################

    #get the function parameters
    my ($arg_ref) = @_;

    my $unified_seq_sample  = $arg_ref->{unified_seq_sample};
    my $trim                = $arg_ref->{trim};
    my $output_dir          = $arg_ref->{output_dir} || '.';
    my $verbose             = $arg_ref->{verbose} || 0;
    my $maximum_lines       = $arg_ref->{number_of_maximum_lines};

    confess "take_snapshot error\n" 
        if any {not defined $_} $unified_seq_sample;

      #set variables:
    my %similar;
    if (defined $maximum_lines) {
        keys %similar =  $maximum_lines;
        #define size of hash to avoid rehasing.
    }
            #similar is used to collpase similar sequence entries
            #it holds a list of cluster positions for each sequence

    my $output_filename = "$output_dir/$D_NAME_OF_UNIFIED_SAMPLE_SEQ_FOR_BLAT";
    my $unique_seq = 0;
            #number of unique sequences for blat
    my $blat_sequences = 0;
            #after breaking very long headers - how many actual sequences we need to blat.

    #prepare the outfile that contains the sequences for training

    print "Collapsing sequences...\n"
        if ($verbose);

    READ_SEQUENCES: 
    {
        my $count = 0;
        open my $UNI_SEQ, '<', $unified_seq_sample 
                    or confess "can't open $unified_seq_sample: $!\n";
        while (<$UNI_SEQ>) {
            chomp;
            my ($lane,$tile,$x,$y,$seq) = split /\s+/ , $_;

            $count++;
            my $header = join '_', ($lane,$tile,$x,$y);


            if (defined $trim) {
                $seq = substr $seq, 0, $trim;
            }
            push @{$similar{$seq}} , $header;
        } # end of READ_SEQUENCES
        close $UNI_SEQ;

#        print "Found $count lines without ambiguous bases ('N's or '.')\n"
#            if ($verbose);

    }

    WRITE_SEQUENCE: 
    {
        open my $FA,'>', $output_filename 
                or confess "can't open $output_filename for writing: $!\n";
        foreach my $seq (keys %similar) { 

            my $collapsed_header = ';';
            $unique_seq++;

            TAKE_CARE_TO_LONG_HEADERS:
            while (scalar @{$similar{$seq}}>0) {

                my $header = shift @{$similar{$seq}};
                my $flag_print = 0;
                my $l_c = length ($collapsed_header) || 0;


                if ($l_c + length ($header) >= $BLAT_MAXIMUM_HEADER_LINE) {
                    $flag_print = 1;
                    push @{$similar{$seq}}, $header;
                }
                elsif (scalar @{$similar{$seq}} == 0) {
                    $flag_print = 1;
                    $collapsed_header = $collapsed_header . ';' . $header;
                }
                else {
                    $collapsed_header = $collapsed_header . ';' . $header;
                }


                if ($flag_print == 1) {
                    $flag_print = 0;
                    $collapsed_header = substr $collapsed_header,2;


                    print {$FA} ">$collapsed_header\n","$seq\n" 
                        or confess "can print into $output_filename: $!\n";
                    #output file in fasta format
                    $blat_sequences++;
                    $collapsed_header = ';';
                }

        }

   } 
        close $FA
            or confess "cant close $output_filename: $!\n";
    } #end of WRITE_SEQUENCE


    print "After processing found ",$unique_seq," unique entries for blat\n",
          "Total blat enteries: ",$blat_sequences,"\n"
        if ($verbose);

    return ($output_filename, $unique_seq);
}







############################# get_filename_in_illumina_format #################################################################################




sub get_filename_in_illumina_format {
    ###############
    #this function generates filenames in solexa format. For example s_4_0002_seq.txt
    #it get in hash table the following keys: path, lane, tile, suffix and symbol.
    #lane and tile are required, default path is './' , default symbol is '_'.
    #path does not need '/' at the end, and symbol is used to concatanate. 
    #for example: get_filename_in_illumina_format({lane => 3, tile => 100, suffix => seq.txt}) 
    #will give ./s_3_0100_seq.txt
    ###############

    my ($arg_ref) = @_;

    my $path      = $arg_ref->{path} || '.';
    my $lane      = $arg_ref->{lane};
    my $r_tiles   = $arg_ref->{tiles};
    my $suffix    = $arg_ref->{suffix};
    my $symbol    = $arg_ref->{symbol} || '_';

    my @results;

    confess 'get_filename_in_illumina_format error\n'
        if any {not defined $_} $lane, $r_tiles ;
        #die if user didn't pass lane or tile

    confess 'get_filename_in_illumina_format error - tiles should be array reference\n'
        if (ref $r_tiles ne 'ARRAY');

    foreach my $tile (@$r_tiles) {
        #generating /path/s_<lane>_000<tile>_<suffix>. like ./s_8_0001_seq.txt
        my $tile_with_zeros = sprintf "%04d" , $tile ;
        my $file = join $symbol, 's', $lane, $tile_with_zeros;
        if (defined $suffix) {
            $file = join $symbol, $file, $suffix;
        }

        my $file_full = join '/', $path, $file;
        push @results, $file_full;
    }
    return @results;
}

# 
# sub any (&@) {
#     my $f = shift;
#     return if ! @_;
#     for (@_) {
# 	return 1 if $f->();
#     }
#     return 0;
# }


################################## test and a template #####################################################

sub test {

    my @tiles = (1..4);
    my $verbose = 1;
    my $path_to_seq_files = '/mnt/mountPoint/usb1/071215_BILLIEHOLIDAY_0001_FC2020WAAXX/Data/C1-50_Firecrest1.8.28_31-12-2007_sbsuser/Bustard1.8.28_31-12-2007_sbsuser';
    my $path_to_int_files = '/mnt/mountPoint/usb1/071215_BILLIEHOLIDAY_0001_FC2020WAAXX/Data/C1-50_Firecrest1.8.28_31-12-2007_sbsuser';
    my $trim = 25;
    my $output_dir = '.';
    my $lines_back = 10000;
    my $lane = 8;

   my ($uni_seq_to_blat_file,$uni_int_file)=quick({verbose=>$verbose,
                                                   path_to_seq_files=>$path_to_seq_files,
                                                   path_to_int_files=>$path_to_int_files,
                                                   trim=>$trim,
                                                   output_dir=>$output_dir,
                                                   maximum_lines_for_training=>$lines_back,
                                                   lane=>$lane,
                                                   tiles=>\@tiles,
                                                   });

}


#####################################################################################################


1;


