#! /usr/bin/env perl

#Copyright 2008 Cold Spring Harbor Laboratory.
# Written by Yaniv Erlich (erlich@cshl.edu)


use strict;
use warnings;

use FindBin qw($Bin);
use lib $Bin;
use lib "$Bin/Cluster";


use Carp;
use Readonly;
use File::Basename;
use Cwd qw(abs_path);
use Getopt::Long;
use Pod::Usage;
use Cluster::Ipar_formatter;

my $VERSION = 0.2;



#^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

my $r_config = read_parameters();
Cluster::Ipar_formatter::quick_formatter($r_config);


#######################################################################################################################
sub test {

    my @tiles = (1..100);
    my @lanes = (3..3);
    quick_formatter ({
                        path_to_ipar_files => '/data/hannon/erlich/Alta-Cyclic/long_runs/path_to_nean',
                        path_to_pos_files  => '/data/hannon/erlich/Alta-Cyclic/long_runs/path_to_nean/pos',
                        r_tiles           => \@tiles,
                        r_lanes           => \@lanes,
                        zipped            => 1,
                        verbose           => 1,
                        debug             => 1,
                        output_dir        => "./",
                    });

}

sub read_parameters {
    my $path_to_ipar_files;
    my $path_to_pos_files;
    my $lanes;
    my $tiles;
    my $zipped;
    my $verbose;
    my $debug;
    my $output_dir;

    my $HELP = 0;
    my $MANUAL = 0;
    my $version = 0;


    my %config;

    GetOptions( 

                "path_to_ipar_files=s"              => \$path_to_ipar_files,
                "path_to_pos_files=s"               => \$path_to_pos_files,
                "lanes=s"                           => \$lanes,
                "tiles=s"                           => \$tiles,
                "verbose"                           => \$verbose,
                "debug"                             => \$debug,
                "zipped"                            => \$zipped,
                "output_dir=s"                      => \$output_dir,

                "help"                              =>\$HELP,
                "man"                               =>\$MANUAL,
                "version"                           =>\$version,
                );

    pod2usage(1) if $HELP;
    pod2usage(-verbose => 2 ) if $MANUAL;

    if ($version) {
        print "version is $VERSION\n";
        exit(1);
    }

    if (
        (not defined $path_to_ipar_files) or
        (not defined $path_to_pos_files)  or
        (not defined $lanes)              or
        (not defined $tiles)              or
        (not defined $output_dir)

       )                                        {
        print "Not all mandatory parameters were loaded\n";
        pod2usage(1);

      }


    #lanes:
    my ($start_lane,$stop_lane) = split /-/ , $lanes;
    if (not defined $stop_lane) {$stop_lane = $start_lane};
    my @lanes = ($start_lane..$stop_lane);

    #tiles:
    my ($start_tile,$stop_tile) = split /-/ , $tiles;
    if (not defined $stop_tile) {$stop_tile = $start_tile};
    my @tiles = ($start_tile..$stop_tile);


    #make sure that directories exist and make path abs.
    if (not -d $path_to_ipar_files) {confess "can't find path to ipar files: $path_to_ipar_files"};
    if (not -d $path_to_pos_files)  {confess "can't find path to pos files: $path_to_pos_files"};
    if (not -d $output_dir)         {confess "can't find output directory: $output_dir"};

    $path_to_ipar_files = abs_path $path_to_ipar_files;
    $path_to_pos_files  = abs_path $path_to_pos_files;
    $output_dir         = abs_path $output_dir;


    #load the parameters:
    $config{'path_to_ipar_files'} = $path_to_ipar_files;
    $config{'path_to_pos_files'}  = $path_to_pos_files;
    $config{'r_lanes'}            = \@lanes;
    $config{'r_tiles'}            = \@tiles;
    $config{'verbose'}            = $verbose;
    $config{'debug'}              = $debug;
    $config{'zipped'}             = $zipped;
    $config{'output_dir'}         = $output_dir;

    return \%config;


}




__END__

=head1 NAME


ipar_formatter - Convert IPAR files to the old and good format of Solexa intensity files.

=head1 SYNOPSIS

ipar_formatter [MANDATORY PARAMETERS]   [OPTIONAL PARAMETERS] 

=cut


=head1 OPTIONS


  Mandatory Parameters:
  --path_to_ipar_files              Path to the folder that contains the ipar files to convert
  --path_to_pos_files               Path to the folder that contains the pos files
  --lanes                           Lanes to convert
  --tiles                           Tiles to convert
  --output_dir                      Path to the target folder. MUST EXISTS before launching the program.

  Optional Parameters:

  --verbose                         Set verbosity.
  --zipped                          If the ipar files are zipped (mostly they are zipped).
  --debug                           Give extra messages and indicators. Stores .o files that return from the cluster

  Help Options:
   --help                           Show a list of parameters.
   --man                            Print a long manual.
   --version                        Print version and exit


=cut


=head1 EXAMPLES

ipar_formatter.pl --path_to_pos_files /data/hannon/erlich/Alta-Cyclic/long_runs/path_to_nean/pos --path_to_ipar_files /data/hannon/erlich/Alta-Cyclic/long_runs/path_to_nean --tiles 1 --lanes 3 --zipped --debug --verbose 
--output_dir .
=cut


=head1 DESCRIPTION

ipar_formatter.pl is part of Alta-Cyclic base caller. The program merages the ipar files (for instance: s_3_0001_int.txt.p.gz) with the pos files (for instance: s_3_0001_pos.txt) to create the format of the intensity files that 
brave_heart and karma likes. It does it in parallel fashion over SGE. 


=cut

=head1 BUGS AND LIMITATIONS

If you feel you have found a bug in ipar_formatter.pl, please submit a bug report at erlich@cshl.edu

=cut

=head1 AUTHOR

Written By Yaniv Erlich <erlich@cshl.edu>

=cut

=head1 LISCENSE AND COPYRIGHT

Copyright © 2008 Cold Spring Harbor Laboratory.
This is free software.  You may redistribute copies of it under the terms in LICENSE file.  There is NO WARRANTY, to the extent permitted by law.

=cut
